package knosso;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Stack;
import java.util.LinkedList;
import java.util.Collections;
import javax.swing.JPanel;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.Timer;

/**
 * Title:        Cnosso
 * Description:  SImulazione di un labirinto
 * Copyright:    Copyright (c) 2001
 * Company:      Universit di Milano
 * @author Sara
 * @version 1.0
 */

public class Maze extends JPanel implements MouseListener,KeyListener
{

  private int hSize = 50;
  private int vSize = 50;

  private Cell[][] mazeMatrix;
  private Stack backtrack;
  private LinkedList neighbors;

  int numCells;
  Cell current;

  int posX=0,posY=0;
  int destX=0,destY=0;

  Image offscreen;
  boolean redraw = true;

  Timer t = new Timer( 5 , new ActionListener() {
      public void actionPerformed( ActionEvent e ) {
      if ((x == destX) && (y == destY)) {
        t.stop();
        return;
      }
       neighbors.clear();
       current.visited = true;
       current.solution = true;

       if( !current.walls[0] && !mazeMatrix[current.posY-1][current.posX].visited)
          neighbors.add(mazeMatrix[current.posY-1][current.posX]);
      if(current.posY < vSize-1)
        if( !current.walls[1] && !mazeMatrix[current.posY+1][current.posX].visited)
          neighbors.add(mazeMatrix[current.posY+1][current.posX]);
      if(current.posX < hSize-1)
        if( !current.walls[2] && !mazeMatrix[current.posY][current.posX+1].visited)
          neighbors.add(mazeMatrix[current.posY][current.posX+1]);
      if(current.posX > 0)
        if(  !current.walls[3] &&  !mazeMatrix[current.posY][current.posX-1].visited)
        neighbors.add(mazeMatrix[current.posY][current.posX-1]);

      if( neighbors.size() > 0 ) {
       Collections.shuffle(neighbors);
       Cell temp = (Cell) neighbors.getFirst();

        backtrack.push( current );
        current = temp;
      } else {
        current.solution = false;
        current = (Cell) backtrack.pop();
      }
      x = current.posX; y = current.posY;
      redraw = true;
      repaint();
    }});



  public Maze() {
    generate();
    addKeyListener(this);
    addMouseListener(this);
    requestFocus();
  }

  public Maze( int vSize , int hSize ) {
    this.hSize = hSize;
    this.vSize = vSize;
    generate();
    addKeyListener(this);
    requestFocus();
  }

  public void setGrid( int x , int y ) {
    vSize = y;
    hSize = x;
  }

  public void generate() {
    // Random numbers for cell selection
    int curX,curY;
    posX = 0 ; posY = 0;
    redraw = true;

     mazeMatrix = new Cell[vSize][hSize];
     backtrack = new Stack();
     neighbors = new LinkedList();

     numCells = vSize*hSize-1;

     // Setup cells
     for( int i=0 ; i < vSize ; i++ )
      for( int j=0 ; j < hSize ; j++ ) {
        mazeMatrix[i][j] = new Cell( j , i );

        // Set borders
        if( i == 0 )
          mazeMatrix[i][j].borders[0] = true;
        else if( i == vSize-1 )
          mazeMatrix[i][j].borders[1] = true;
        else if( j == hSize )
          mazeMatrix[i][j].borders[2] = true;
        else if( j == 0 )
          mazeMatrix[i][j].borders[3] = true;
      }


     // Randomly set starting cell
     curX = Math.round((float) Math.random()*(hSize-1));
     curY = Math.round((float) Math.random()*(vSize-1));

     // Randomly set destination cell
     destX = Math.round((float) Math.random()*(hSize-1));
     destY = Math.round((float) Math.random()*(vSize-1));

     current = mazeMatrix[curY][curX];
     // Generation cycle (DFS)
     while( numCells > 0 ) {
       neighbors.clear();
       current.visited = true;

       if( !current.borders[0] && !mazeMatrix[current.posY-1][current.posX].visited)
          neighbors.add(mazeMatrix[current.posY-1][current.posX]);
      if(current.posY < vSize-1)
        if( /* !current.borders[1] && */ !mazeMatrix[current.posY+1][current.posX].visited)
          neighbors.add(mazeMatrix[current.posY+1][current.posX]);
      if(current.posX < hSize-1)
        if( /* !current.borders[2] && */ !mazeMatrix[current.posY][current.posX+1].visited)
          neighbors.add(mazeMatrix[current.posY][current.posX+1]);
      if(current.posX > 0)
        if( /* !current.borders[3] && */ !mazeMatrix[current.posY][current.posX-1].visited)
        neighbors.add(mazeMatrix[current.posY][current.posX-1]);

      if( neighbors.size() > 0 ) {
       Collections.shuffle(neighbors);
       Cell temp = (Cell) neighbors.getFirst();

        if( temp.posY == current.posY-1 ) {
          current.walls[0] = false;
          temp.walls[1] = false;
        } else if( temp.posY == current.posY+1 ) {
          current.walls[1] = false;
          temp.walls[0] = false;
        } else if( temp.posX == current.posX+1 ) {
          current.walls[2] = false;
          temp.walls[3] = false;
        } else if( temp.posX == current.posX-1 ) {
          current.walls[3] = false;
          temp.walls[2] = false;
        }

        backtrack.push( current );
        current = temp;
        numCells--;
      } else {

        current = (Cell) backtrack.pop();

      }
    }
    for( int i=0 ; i < vSize ; i++ )
      for( int j=0 ; j < hSize ; j++ )
        mazeMatrix[i][j].visited = false;
    repaint();

  }

  public void drawMaze() {
    int blockSizeX = getSize().width/hSize;
    int blockSizeY = getSize().height/vSize;

//    if(offscreen == null) {
    offscreen = createImage(getSize().width, getSize().height);
//    }

    Graphics og = offscreen.getGraphics();
    og.setClip(0,0,getSize().width, getSize().height);

    for( int i=0 ; i < vSize ; i++ ) {
      for( int j=0 ; j < hSize ; j++ ) {
         if( mazeMatrix[i][j].visited ) {
          og.setColor( Color.orange );
          og.fillRect( j*blockSizeX , i*blockSizeY ,blockSizeX, blockSizeY );
          og.setColor( Color.black );
        }
        if( mazeMatrix[i][j].solution ) {
          og.setColor( Color.yellow );
          og.fillRect( j*blockSizeX , i*blockSizeY ,blockSizeX, blockSizeY );
          og.setColor( Color.black );
        }
        if( mazeMatrix[i][j].posX == destX && mazeMatrix[i][j].posY == destY ) {
          og.setColor( Color.red );
          og.fillRect( j*blockSizeX , i*blockSizeY ,blockSizeX, blockSizeY );
          og.setColor( Color.black );
        }
        if( mazeMatrix[i][j].posX == posX && mazeMatrix[i][j].posY == posY ) {
          og.setColor( Color.green );
          og.fillRect( j*blockSizeX , i*blockSizeY ,blockSizeX, blockSizeY );
          og.setColor( Color.black );
        }
        if( mazeMatrix[i][j].walls[0] )
          og.drawLine( j*blockSizeX , i*blockSizeY ,(j+1)*blockSizeX, i*blockSizeY );
        if( mazeMatrix[i][j].walls[1] )
          og.drawLine( j*blockSizeX , (i+1)*blockSizeY ,(j+1)*blockSizeX, (i+1)*blockSizeY );
        if( mazeMatrix[i][j].walls[2] )
          og.drawLine( (j+1)*blockSizeX , i*blockSizeY ,(j+1)*blockSizeX, (i+1)*blockSizeY );
        if( mazeMatrix[i][j].walls[3] )
          og.drawLine( j*blockSizeX , i*blockSizeY , j*blockSizeX, (i+1)*blockSizeY );
      }
      og.dispose();
    }

  }

  public void drawBox( Graphics og , int j , int i ) {
    int blockSizeX = getSize().width/hSize;
    int blockSizeY = getSize().height/vSize;

    if( mazeMatrix[i][j].walls[0] )
      og.drawLine( j*blockSizeX , i*blockSizeY ,(j+1)*blockSizeX, i*blockSizeY );
    if( mazeMatrix[i][j].walls[1] )
      og.drawLine( j*blockSizeX , (i+1)*blockSizeY ,(j+1)*blockSizeX, (i+1)*blockSizeY );
    if( mazeMatrix[i][j].walls[2] )
      og.drawLine( (j+1)*blockSizeX , i*blockSizeY ,(j+1)*blockSizeX, (i+1)*blockSizeY );
    if( mazeMatrix[i][j].walls[3] )
      og.drawLine( j*blockSizeX , i*blockSizeY , j*blockSizeX, (i+1)*blockSizeY );
  }

  public void printMaze() {
    for( int i=0 ; i < vSize ; i++ ) {
      for( int j=0 ; j < hSize ; j++ ) {
        if( mazeMatrix[i][j].walls[1] && mazeMatrix[i][j].walls[3] )
          System.out.print("_|");
        else if( mazeMatrix[i][j].walls[1] )
          System.out.print("_ ");
        else if( mazeMatrix[i][j].walls[3] )
          System.out.print(" |");
        else System.out.print("  ");
      }
      System.out.print("\n");
    }
  }

  int x=0,y=0;
  boolean trace = true;

  public void setTrace() {
    trace = !trace;
  }

  public void findSolution() {
    x=0;y=0;

    backtrack.clear();
    current = mazeMatrix[y][x];

    for( int i=0 ; i < vSize ; i++ )
      for( int j=0 ; j < hSize ; j++ )
        mazeMatrix[i][j].visited = false;

     // Solution cycle (DFS)

    if( trace ) {
      t.start();
    } else {
     while( (x != destX) || (y != destY)) {
       neighbors.clear();
       current.visited = true;
       current.solution = true;

       if( !current.walls[0] && !mazeMatrix[current.posY-1][current.posX].visited)
          neighbors.add(mazeMatrix[current.posY-1][current.posX]);
      if(current.posY < vSize-1)
        if( !current.walls[1] && !mazeMatrix[current.posY+1][current.posX].visited)
          neighbors.add(mazeMatrix[current.posY+1][current.posX]);
      if(current.posX < hSize-1)
        if( !current.walls[2] && !mazeMatrix[current.posY][current.posX+1].visited)
          neighbors.add(mazeMatrix[current.posY][current.posX+1]);
      if(current.posX > 0)
        if(  !current.walls[3] &&  !mazeMatrix[current.posY][current.posX-1].visited)
        neighbors.add(mazeMatrix[current.posY][current.posX-1]);

      if( neighbors.size() > 0 ) {
       Collections.shuffle(neighbors);
       Cell temp = (Cell) neighbors.getFirst();

        backtrack.push( current );
        current = temp;
      } else {
        current.solution = false;
        current = (Cell) backtrack.pop();
      }
      x = current.posX; y = current.posY;
      }
      redraw = true;
      repaint();
    }
  }

  public void clearSolution() {
    t.stop();
    for( int i=0 ; i < vSize ; i++ )
      for( int j=0 ; j < hSize ; j++ ) {
        mazeMatrix[i][j].solution = false;
        mazeMatrix[i][j].visited = false;
      }
    posX = 0; posY = 0;
    redraw = true;
  }

  public void save() {
    String temp = null;
    JFileChooser chooser = new JFileChooser();
    int returnVal = chooser.showSaveDialog(null);
    if(returnVal == JFileChooser.APPROVE_OPTION)
      temp =  chooser.getSelectedFile().getPath();
    else return;
    try {
      FileOutputStream ostream = new FileOutputStream(temp);
      ObjectOutputStream p = new ObjectOutputStream(ostream);

      p.writeObject(mazeMatrix);
      p.writeInt(destX);
      p.writeInt(destY);
      p.writeInt(posX);
      p.writeInt(posY);

      p.flush();
      ostream.close();

    } catch ( IOException ex ) {
      System.out.println( "Problema con i file di labirinto" );
    }
  }

 public void load() {
    String tmp = null;
    JFileChooser chooser = new JFileChooser();
    int returnVal = chooser.showOpenDialog(null);
    if(returnVal == JFileChooser.APPROVE_OPTION)
        tmp =  chooser.getSelectedFile().getPath();
    else return;
      try {
        FileInputStream istream = new FileInputStream(tmp);
        ObjectInputStream p = new ObjectInputStream(istream);

        mazeMatrix = (Cell[][]) p.readObject();
        destX = p.readInt();
        destY = p.readInt();
        posX = p.readInt();
        posY = p.readInt();

        istream.close();

      } catch ( Exception ex ) {
        System.out.println( "Problema con i file di labirinto" );
      }
      redraw = true;
      repaint();
  }

  public void paint(Graphics g) {
    int blockSizeX = getSize().width/hSize;
    int blockSizeY = getSize().height/vSize;

    if( redraw ) {
      drawMaze();
      redraw = false;
    }

    Graphics og = offscreen.getGraphics();
    og.setClip(0,0,getSize().width, getSize().height);

    og.setColor( Color.red );
    og.fillRect( destX*blockSizeX , destY*blockSizeY ,blockSizeX, blockSizeY );
    og.setColor( Color.black );
    drawBox( og , destX , destY );

    og.setColor( Color.green );
    og.fillRect( posX*blockSizeX , posY*blockSizeY ,blockSizeX, blockSizeY );
    og.setColor( Color.black );
    drawBox( og , posX , posY );

    g.drawImage(offscreen, 0, 0, null);
    //og.clearRect( 0 , 0 , getSize().width , getSize().height);
    og.dispose();

  }

  public void keyPressed( KeyEvent e ) {
  }
  public void keyReleased( KeyEvent e ) {
    int blockSizeX = getSize().width/hSize;
    int blockSizeY = getSize().height/vSize;
    mazeMatrix[posY][posX].solution = true;

    Graphics og = offscreen.getGraphics();
    og.setClip(0,0,getSize().width, getSize().height);
    og.setColor( Color.yellow );
    og.fillRect( posX*blockSizeX , posY*blockSizeY ,blockSizeX, blockSizeY );
    og.setColor( Color.black );
    drawBox( og , posX , posY );
    og.dispose();

    switch( e.getKeyCode() ) {
      case KeyEvent.VK_UP :
        //posY>0?posY--:posY=0;
        if( !mazeMatrix[posY][posX].walls[0] )
          posY--;
        break;
      case KeyEvent.VK_DOWN :
        //posY<VSIZE?posY++:posX=VSIZE;
        if( !mazeMatrix[posY][posX].walls[1] )
        posY++;
        break;
      case KeyEvent.VK_LEFT :
        //posX>0?posX--:posX=0;
        if( !mazeMatrix[posY][posX].walls[3] )
        posX--;
        break;
      case KeyEvent.VK_RIGHT :
        //posX<HSIZE?posX++:posX=HSIZE;
        if( !mazeMatrix[posY][posX].walls[2] )
        posX++;
        break;
      default:
    }
    if( posX == destX && posY == destY )
      JOptionPane.showMessageDialog( this ,
                                   "Hai trovato il minotauro!",
                                   "Vittoria!",
                                   JOptionPane.WARNING_MESSAGE);
    repaint();
  }
  public void keyTyped( KeyEvent e ) {
  }

  public void mouseClicked( MouseEvent e ) {
    requestFocus();
  }
  public void mousePressed( MouseEvent e ) {
  }
  public void mouseReleased( MouseEvent e ) {
  }
  public void mouseEntered( MouseEvent e ) {
  }
  public void mouseExited( MouseEvent e ) {
  }

}

class Cell implements java.io.Serializable {
  // Walls               N     S       E     W
  boolean walls[] = { true , true , true , true };

  // Borders
  boolean borders[] = { false , false , false , false };

  // Visited flag
  boolean visited = false;

  // Solution flag
  boolean solution = false;

  // Position
  int posX,posY;

  public Cell( int posX , int posY ) {
    this.posX = posX;
    this.posY = posY;
  }

  public String toString() {
    return posX+"--"+posY;
  }

}
